* 64K address space minimal fig forth system *
 NOLIST
        section forth   for the gst-macro-assembler
        pagewid 120
        pagelen 20000   (no form feed)

* assembler:
* strings/numbers/-expressions evaluated and assembled in [] brackets
* equates as labels by their names (as usual)
* n1<<n2 :  shift n1 the no. of n2 places to the left
* n1!n2  :  logical or
* n1&n2  :  logical and
* .len(variable) : evaluates the length of the chars of the vari
* .left(variable,n1) : the leftmost n1 chars of the vari
* .right(...)              right...
*
* variables:
* name setstr [variable] or [expression] or {characters}   string
* name setnum [variable] or number                         number
*
*    dc.b    ''''       assembles one single quote as string data
*    '`' instead of '[' to substitute the left bracket which this
*                       assembler wouldn't accept as string data
*
* create macro t,xaddr          defines a macro:
* name         parameters       parameters being evaluated as variables
* [.lab]                        insert the call label
* name  maclab                  local macro label

* envirenmental depencies can be found looking for the text: "qdos:"
* trap #0
*   used to enter the supervisor state
* trap #1
*   system management
* trap #2
*   i/o allocation
* trap #3
*   i/o operations
*   all i/o operations expect the channel identifier in a0.l
*   the opcode in d0.l(.b) and the timeout in d3.w (-1 forever, +ve * 20 ms)
* qdos-strings
*   are always preceeded by a count-word unless stated
* trap #4
*   preceeding a call to trap #2 or trap #3 the latter would
*   access a0.l (trap#2) or a1 (trap#3) offset by a6.l
* trap #5 to trap #15
*   user definable

;;;dataspace equ 4000            enough to be getting on with!
;;;        data dataspace

* Registers: (D)estroyable, (I)nput only, (F)orth reserved
     ; d0-d3 (D)
     ; d4-d6 (D) never touched! user code can communicate
     ;    d7.msw (D) zeroed at boot, cold or warm starts
up      setstr  d7  lsw  (f) user pointer, needing offset by a6
     ;          a0       (d) destroyabel, used for channel-id's
t       setstr  a1       (i) absolute address of entered code
w       setstr  a2       (i) parameter field pointer
nx      setstr  a3       (f) jump address to leave assembler
ip      setstr  a4       (f) forth next instruction pointer
rp      setstr  a7 <=>a5 (f) return stack pointer
bp      setstr  a6 <=> Z (f) always the program base address
sp      setstr  a5 <=>a7 (f) forth stack pointer, params in/out

* ';;;' code corrected/changed for security
 LIST
disp    equ     $8000-104       64K address space, the job header inclusive
*
chans   equ     16              number of channels supported (must be even)
*
ssbits  equ     6               is nice for 64 byte headers
ssize   equ     1<<ssbits
nbuf    equ     1024/ssize
*
bmag    equ     (ssize+4)*nbuf  buffer space
uarea   equ     128 ;64   user area
tibrp   equ     512 ;256  combined tib & return stack

 NOLIST
* macros
*
*;;*prevlink setstr Z+$8000   ;;Z
prevnum setnum  0
previmm setnum  0
sings   setstr  {'`Xx}

voc     setstr  forth   ;;
lastvoc setstr  Z+$8000 ;;
forthlk setstr  Z+$8000 ;;
rootlk  setstr  Z+$8000 ;;
*
create  macro   t xaddr
z[prevnum] equ  [previmm]
l       setnum  [.len(t)]
prevnum setnum  [prevnum]+1
        dcb.b   (*+1+[l]-Z)&1+1,z[prevnum]![l] for cet, etc
        ifnum   [l]             = 1 goto single
        ifstr   {[.left(t,1)]} <> {`} goto ord
        dc.b    $5B,'compile',']'!$80
        goto    dclfa
single  maclab  
        ifnum   [.instr(sings,t)] = 0 goto nsing
        ifstr   {[t]} <> ' goto ntick
        dc.b    ''''!$80
        goto    dclfa
ntick   maclab  
        ifstr   {[t]} <> {`} goto nbrac
        dc.b    91!128
        goto    dclfa
nbrac   maclab  
        dc.b    -128
        goto    dclfa
nsing   maclab  
        dc.b    '[t]'!$80
        goto    dclfa
ord     maclab  
        dc.b    '[.left(t,[l]-1)]','[.right(t,1)]'!$80
dclfa   maclab  
        dc.w    [[voc]lk]-Z ;;prevlink]-Z
[.lab]  dc.w    [xaddr]-Z
prevlab setstr  [.lab]
[voc]lk setstr [.lab]-[l]-3
lastlk  setstr [[voc]lk]
previmm setnum  128
        endm

* example:
*voc setstr root
*newlabel create {forthname} cfalabel
* then
*[voc]lk setstr [.lab]-[l]-3
* expands to "rootlk" as a text variable set to
*            "newlabel-9-3"
*[[voc]lk] => [rootlk] => the label just derived

immed   macro   
previmm setnum  $140-[previmm]   {320}
        endm    
*
;endit   macro
;z[prevnum] equ  [previmm]
;ntop    equ     prevlink]
;top     ds.w    0
;        end
;        endm

*
* standard layout for QDOS, and me
*
 LIST
Z equ *+disp
go      lea     Z(pc),a6        ;;;for the hotkey-"system" which
        bra.s   Zentr           ;;;on job entry changes the regs
*** qdos job-header
        dc.w    $4AFB
qfn equ *+1
        dc.w    2,'Q6'          to lengthen, patch 'screen' to elsewhere.
screen  dc.w    $201,$207
scsiz   dc.w    448,202,32,14   border colr & width, paper, ink; window
zsavehead ; at locn 24 (!)
        dc.l    0               24 file length, set to 'here' in case save" ser" used
        dc.b    0,1             28 access code = 0, filetype = executeable
        dc.l    dataspace,0     30,34
* 30.l/32.w: dataspace, may be altered, but keep <= 32K-'here'!
* 
        dc.b    chans-2         38
ctemp   dc.b    chans-1         39
cstdio  dc.b    0,1,2,3         40..43
 NOLIST
* The above give a simple redirection of our channel table.
* 38: 14 dr0 (r/w, etc)
* 39: 15 dr1 (ditto, plus save" and chain")
* 40:  0 standard 1st EX-chan or CON, input (expect, key, ?terminal, etc)
* 41:  1 standard 2nd or only EX-chan or CON, output (type, emit, etc)
* 42:  2 reserved not used, for user code, e.g. as an input or update file
* 43:  3 reserved not used, for user code, e.g. as an output or work file
* channels 0,1,3-11: may be passed in by EX
*             12-13: uncommitted
*
* boot up code
* qdos:
* starting a qdos job by        EX jobfilename,#channel1,..;commandstring
* bot|A6           |(A4,A6)              ..(a7)..    |(A5,A6) top
*    |codespace    |dataspace   initial sp:parameters|
* optional parameters on stack: count.w|qdos-channels.l | count.w|command stg|
*
 LIST
Zentr   move.w  (A7)+,d7
        bne.s   gotcha          ?
defscr  lea     screen-Z(a6),a1 default window colours & sizes
        move.w  198,a2          qdos vector
        jsr     (a2)            > channel open and window set up routine
        move.l  a0,-(A7)        put default screen on stack
        moveq   #1,d7           say it's there
gotcha  movem.l (A7),d0-d1/d3-d6/a0-a4
        lsl.w   #2,d7
        add.w   d7,A7           top of channels
        lea     -chans*4(A7),a5 channel table base
        subq.w  #4,d7           did we just have one channel?
        bne.s   notone          ?
        move.l  d0,d1           yes - make ch1=ch0
        addq.w  #4,d7           and pretend it was there all along
notone  addq.w  #4*2,d7         put back and include channel 2
        moveq   #-1,d2          channel 2 is initially closed
        movem.l d0-d6/a0-a4,(a5) set down what seems to be going
        move.l  a7,a4   ;;for the command stg
        add.l   a5,d7
chclear move.l  d2,-(A7)        set remaining channels closed
        cmp.l   A7,d7
        bne.s   chclear         ?
;;* safe address range:
;        lea     32768-chans*4-4(a6),a1 max limit
;        move.l  a5,d1
;        sub.l   a1,d1
;        bls.s   toptop  ? valid address range
;        moveq   #1,d0   cmd.count - loopindex + odd byte
;        add.w   (a7),d0 cmd.length
;        add.l   d0,d1   channels
;        move.l  a1,a5
;        add.l   a1,d1
;topmov  move.w  (a7)+,(a1)+
;        cmp.l   a1,d1
;        bhi.s   topmov
toptop  move.l  a5,a0
        sub.l   a6,a5
        move.w  a5,limit+2-Z(a6)
        lea     -bmag(a5),a1
        move.w  a1,first+2-Z(a6)
        move.w  a1,prev+2-Z(a6)
        move.w  a1,use+2-Z(a6)
        lea     -uarea(a1),[rp] r0
        move.w  [rp],[up]
        lea     -tibrp([rp]),a1 tib
        move.w  a1,[sp]         s0
        exg     [rp],[sp]
        movem.w [up]/a1/[rp]/[sp],origin+16-Z(a6)
        exg     [rp],[sp]       => A5:=SP A7=:RP
*
        add.l   a6,[sp]
btclr   clr.l   -(a0)
        cmp.l   [sp],a0
        bcc.s   btclr           ?
* commandstring to tib:
        move.l  [sp],a1
        move.w  (a4)+,d1
        ble.s   origin  ? no commands
        subq.w  #1,d1
        cmp.w   #tibrp-65,d1
        bgt.s   origin  ? too long
        move.w  #loadquit+2-Z,vpquitq-Z(a6)
cmdnx   move.b  (a4)+,(a1)+
        dbra    d1,cmdnx
* base of fig-FORTH proper
origin  jmp     docold-Z(a6)    00
        jmp     dowarm-Z(a6)    04
        dc.l    $4ef6ec54       08 lwr68k in radix 36 {1324805204}
* cold/warm move the following to the user area
        dc.w    lasttop-Z ;;ntop-Z          12 last name in dictionary
        dc.w    194             14 backspace character (never used)
        ds.w    4               16 initial up, s0, r0 and tib - filled in at prog exec
        dc.w    31              18 initial max width for names
        dc.w    0               20 0: no disk, 1: dr0 available, -1: (abort) in use
* warm stops here
        dc.w    lktop-Z  ;;top-Z       22 initial fence address
        dc.w    dptop    ;;top-Z       24 initial top of directory
        dc.w    voctop-Z ;;vl0-Z       26 initial vocabulary link pointer
* cold goes all the way, and also resets forth+6 from origin+$c

* qdos:
* fetch max. d2.w bytes or until <lf> from channel-id a0.l to  a1.l)+
*            d3.w waiting for(-1)ever, +ve no. of 20ms periods
* flag -ve error code in d0.l, or 0 if ok, -1 time elapsed
xexpect move.w  ([sp])+,d2
        move.w  ([sp])+,a1
        moveq   #2<<3!4!2,d0    io.fline wait stdin
        bsr.s   iocomm          >
        bpl.s   expa1ok         ?
        addq.l  #1,a1
        sf      cstdio-Z(a6)    any error, try resetting to chan 0
expa1ok sf      (a1)
        sf      -(a1)
        jmp     ([nx])

* qdos routine:
* io.cursen     enable cursor of (window-)channel a0
* io.fbyte      fetch one byte from channel-id a0 to d1.b
xkey    moveq   #14<<3!4!2,d0   io.cursen wait stdin
        bsr.s   iocomm          >
        moveq   #1<<3!4!2,d0    io.fbyte wait stdin
        bsr.s   iocomm          >
        move.w  d1,-([sp])
        sf      ([sp])
        jmp     ([nx])

* qdos: io.pend test wether bytes are waiting for beeing fetched
*       flag result d0.l =0 if, d0=-1 time elapsed
xquestterminal
        moveq   #14<<3!0!2,d0   io.cursen nowait stdin
        bsr.s   iocomm          >
        moveq   #0<<3!0!2,d0    io.pend nowait stdin
        bsr.s   iocomm          >
        not.l   d0              c=0, err.nc will give z
        bra   hitrue

doemit  move.w  ([w]),-([sp])
xemit   move.w  ([sp])+,d1
        moveq   #5<<3!4!3,d0    io.sbyte wait stdout
        moveq   #1,d2
        bra.s   iocount

* qdos:
* io.sstrg send a string of < 32K d2.w bytes from a1.l)+ to channel-id a0
xtype   move.w  ([sp])+,d2
        move.w  ([sp])+,a1
        moveq   #7<<3!4!3,d0    io.sstrg wait stdout
iocount add.w   d2,26(a6,[up].w)
        move.l  [nx],-(A7)
iocomm  moveq   #3,d3
        and.b   d0,d3
        lea     -2(a6,d3.w),a0  ;;
        move.b  cstdio-Z(a0),d3 ;;move.b  cstdio-2-Z(a6,d3.w),d3
        bsr.s   mkchp           >
        move.l  d3,a0
        lsr.w   #3,d0
        subx.w  d3,d3
trap3   add.l   a6,a1
        move.l  a1,-([sp])
        trap    #3
        move.l  ([sp])+,a1
        add.w   d1,a1 Ho hum... ERR.BP screws a1, this is for 'expect'
        tst.l   d0
        rts     

xioblo  move.w  ([ip])+,d0
        movem.w ([sp])+,d2/a1
        bsr.s   iocomm          >
        bra.s   rwtst

mkchp   and.w   #15,d3
        lsl.b   #2,d3
        move.w  limit+2-Z(a6),a0
        add.l   a6,a0
        lea     0(a0,d3.w),a2
        move.l  (a2),d3
        rts     

* qdos: fs.posab fetch filepointer of channel-id a0 to d1.l
xrbyw   move.w  ([sp])+,d2
        moveq   #0,d1
        move.w  ([sp])+,d1
        move.w  #66<<3!4!0/2,d0 fs.posab wait dr0
        asl.w   #2,d1           block.b14->x
        addx.w  d0,d0           if bit 14 was set, change to dr1
        lsl.l   #ssbits-2,d1
        bsr.s   iocomm          >
        move.w  ([sp])+,a1
        bmi.s   rwerr           ?
        moveq   #7,d0           io.sstrg
        lsr.b   d2,d0           if flag was 1, convert to 3=io.fstrg
        moveq   #ssize,d2
        bsr.s   trap3           >
rwtst   bpl.s   gonext          ?
rwerr   lea     quit+2-Z(a6),[ip]
        move.l  d0,-([sp])
        move.w  #1,-([sp])
xreport move.w  ([sp])+,d3
        move.l  ([sp])+,d0
        pea     ([nx])
        bsr.s   mkchp   > find id of chan. no. D3
        move.l  d3,a0   window channel-id
        move.w  204,a2  ut_err
        jmp     (a2)    error message

xert    move.l  ([sp])+,d0
        bra.s   rwtst  -

* qdos: close channel identified in a0.l
Xclose  move.w  ([sp])+,d3
        move.l  [nx],-(A7)
closeit bsr.s   mkchp           >
        moveq   #-1,d1          (also for open)
        move.l  d1,(a2)
        moveq   #2,d0           io.close
        cmp.w   d3,d0
        bge.s   noclose         ?ignore -ve and prevent close of qdos 0-2
        moveq   #chans-1,d2
clchk   cmp.l   (a0)+,d3
        dbeq    d2,clchk
        beq.s   noclose         ?prevent close of duplicate entry
        move.l  d3,a0
        trap    #2
noclose rts

* qdos: open a channel to the device with its name at a0.l)+
*       return its identifier in a0.l
xopen   move.w  ([sp])+,d3
        bsr.s   closeit         >(kindly sets d1=-1 for us)
        movem.w ([sp])+,d3/a0
        add.l   a6,a0
        moveq   #1,d0           io.open
        trap    #2
        move.l  d0,-([sp])
        blt.s   gonext          ?
        move.l  a0,(a2)
gonext  jmp     ([nx])

* bye re below, qdos: remove the 4th job
*       close its channels, link its resources back to the system

** ** ** 4th system

docold  lea     Z(pc),a6        ;;
        movem.w origin+12-Z(a6),d0-d3/[up]/a0-a5  d2->UP
        move.w  d0,forth+6-Z(a6)
        movem.w a3-a5,16(a6,d2.w)
dowarm  lea     Z(pc),a6        ;;
        movem.w origin+12-Z(a6),d0-d1/[up]/a0-a4
        movem.w d0/d1/[up]/a0-a4,0(a6,[up].w)
        lea     abort+2-Z(a6),[ip]
        lea     nextloc-Z(a6),[nx]
xrpstore move.w 8(a6,[up].w),[rp]
        add.l   a6,[rp]
        jmp     ([nx])

voc setstr forth ; 1st in the dictionary

voc setstr forth
*root    create  {(root)} dodoesgreat
*        dc.w dovoc-Z
*        dc.w $81a0       ; 1st nfa
*        dc.w roottop-Z         lfa
*rootv   dc.w [lastvoc]-Z ; -> $8000
*lastvoc setstr  rootv
*rootlk  setstr  rootv-4
*previmm setnum  $140-[previmm]

noop create {noop} jnxcode

*voc setstr root
forth   create  {forth} dodoesgreat
        dc.w dovoc-Z     -> pfa from tick
        dc.w $81a0       -> 1st voc.nfa
        dc.w forthtop-Z  -> (current)cfa - (context) ;;ntop-Z
forthv  dc.w [lastvoc]-Z -> (voc-link)
lastvoc setstr  forthv
** **forthlk setstr  forthv-4 ; un-comment if root un-commented
previmm setnum  $140-[previmm]

* qdos: remove the 4th job
*       close its channels, link its resources back to the system
bye     create  {bye} *+2
        moveq   #0,d3   no error, returned to owner job
        moveq   #-1,d1
        moveq   #5,d0
        trap    #1      (terminates the current job)

voc setstr forth

lit     create  {lit} *+2
        move.w  ([ip])+,-([sp])
nextloc move.w  ([ip])+,[w]
exent   add.l   a6,[w]
        move.w  ([w])+,a1
        jmp     0(a6,a1.w) ;;;add.l   a6,a1
                           ;;;jmp     (a1)

dlit    create  {dlit} *+2
        move.l  ([ip])+,-([sp])
jnxcode jmp     ([nx])

execute create  {execute} Xexec
cfaexec dc.w Xcfaex-Z
                ;;perform create  {perform} *+2
                ;;move.w  ([sp])+,[w]
                ;;move.w  0(a6,[w].w),-([sp])
dodefer moveq   #-7,d0  qdos-code err.nf
        move.w  ([w])+,d2
        move.w  udp(a6,[up].w),d1
        cmp.w   d2,d1
        bgt.s   dfdex   ? ok
        move.w  ([w]),d2
        cmp.w   d2,d1
        ble.s   dfder   ? error
        move.w  d2,-([w])
dfdex   move.w  d2,-([sp])
        ror.b   #1,d2
        bcs.s   dfder   ? odd
Xcfaex  subq.w  #2,([sp])
Xexec   move.w  ([sp])+,[w]
        bra.s   exent   -
dfder   bra     rwerr   -

defts   create {defaults} *+2
        move.w  ([sp])+,[w]
        add.l   a6,[w]
        move.b  2([w]),([w])+
        move.b  2([w]),([w])+
        jmp     ([nx])

branch  create  {branch} Xbranch

zerobranch create {0branch} *+2
        tst.w   ([sp])+
        beq.s   xbranch         ?
bump    addq.l  #2,[ip]
        jmp     ([nx])

ploopq  create  {(loop)} *+2
        moveq   #1,d0
loopcmp add.w   ([rp])+,d0
        cmp.w   ([rp]),d0
        bge.s   loopex          ?
loopon  move.w  d0,-([rp])
xbranch add.w   ([ip]),[ip]
        jmp     ([nx])

pplusloopq create {(+loop)} *+2
        move.w  ([sp])+,d0
;;; changed to avoid endless looping on zero increment:
        bgt.s   loopcmp   ;;;bpl.s     loopcmp
        add.w   ([rp])+,d0
        cmp.w   ([rp]),d0
        bgt.s   loopon          ?
loopex  addq.l  #2,[rp]
        bra.s   bump

pdoq    create  {(do)} *+2
        move.l  ([sp])+,-([rp])
        jmp     ([nx])

digit   create  {digit} *+2
        move.w  ([sp])+,d1
        moveq   #-'0',d0
        add.w   ([sp]),d0
        bmi.s   putfalse        ?
        cmp.w   #10,d0
        bcs.s   digbas          ?
        cmp.w   #'a'-'0',d0
        bcs.s   diglow          ?
        sub.w   #'a'-'A',d0
diglow  cmp.w   #'A'-'0',d0
        bcs.s   putfalse        ?
        subq.w  #'A'-'0'-10,d0
digbas  move.w  d0,([sp])
        cmp.w   d1,d0
        bcs   xone            ?
putfalse clr.w  ([sp])
        jmp     ([nx])

* prepare word names to bit 7 unset
hedchars dc.w *+2-Z
        move.l  ([sp]),d1
        lea     0([bp],d1.w),a0
        swap    d1
        subq.w  #1,d1
        bmi.s   nohedc  ?
lphedc  and.b   #$7f,(a0)+
        dbra    d1,lphedc
nohedc  jmp     ([nx])

** uncomment this one, or the other
* (find) ( stg vocnfa -- pfa char tf | ff
pfindq  create  {(find)} *+2
        moveq   #0,d0
        move.w  ([sp])+,[w] top nfa
        moveq   #-1,d1
        move.w  uwidth([bp],[up].l),d3
        subq.w  #1,d3   limit of char count
        swap    d3
        bra.s   findent -
findlk  move.l  d3,d2
        swap    d2
        tst.b   (a0)
        bpl.s   fndfalse ?
        moveq   #31,d0
        and.b   (a0),d0
        beq.s   fndfalse ?
        cmp.w   d0,d2
        bge.s   fndwdt  ? width ok
        move.w  d2,d0
fndwdt  lea     1(a0,d0.w),[w] lfa
findvoc move.l  [w],d1
        ble.s   fndfalse  ? memory corrupt
        lsr.w   #1,d1
        bcs.s   fndfalse  ? no more / address err.
        add.w   d1,d1
        move.w  ([w]),[w] nfa
        cmp.w   [w],d3
        beq.s   fndfalse  ? exit searching loop
findent move.w  [w],d3
        cmp.w   #$8000,d3
        beq.s   fndfalse  ? end/empty
        add.l   [bp],[w]  top.nfa
        move.l  [w],a0
        move.w  ([sp]),a1
        add.l   [bp],a1   strg
        move.b  ([w])+,d0
        cmp.b   #$81,d0
        bne.s   fndtry    ? proceed
        cmp.b   #$a0,([w])
        bne.s   fndtry    ? more in this voc
fndfalse clr.w  ([sp])
        jmp     ([nx])
fndtry  moveq   #$3f,d2   mask off precedence bit
        and.b   d0,d2     smudge & count
        sub.b   (a1)+,d2  len
        bne.s   findlk  ?
findscan moveq  #127,d2
        and.b   ([w]),d2  char
        moveq   #127,d1
        and.b   (a1)+,d1
        eor.b   d2,d1
        beq.s   findchk   ? char =
        cmp.b   #32,d1     u-case
        bne.s   findlk  ? char =/=
        or.b    d1,d2
        sub.b   #'a',d2
        cmp.b   #26,d2
        bcc.s   findlk  ? > 'z'
findchk tst.b   ([w])+
        bpl.s   findscan  ? nxt char
        addq.l  #4,[w]    pfa
        sub.l   [bp],[w]
        move.w  [w],([sp])
        move.w  d0,-([sp])
xone    move.w  #1,-([sp])
        jmp     ([nx])
** uncomment this one, or the other
*pfindq  create  {(find)} *+2
*        moveq   #0,d0
*        move.w  ([sp])+,[w] voc-nfa
*        bra.s   findent
*findlink                ;;tst.b  ([w])+
*                        ;;bpl.s   findlink        ?
*                        ;;move.l  a0,[w]  ;;lfa
*        move.w  (a0),[w]  next link     ;;move.w  ([w]),[w]
*findent                 ;;move.w  a0,d2     ;;
*        move.w  [w],d2  ;;;
*        cmp.w   #$8000,d2 ;
*        beq.s   findnx  ? ;putfalse        ?
*        add.l   a6,[w]
*        move.w  ([sp]),a1
*        add.l   a6,a1
*        move.b  ([w]),d0 ;([w])+,d0
*        move.l  [w],a0  ;;
*        moveq   #31,d2  ;;
*        and.b   d0,d2   ;;
*        beq.s   findnx  ?       ;;putfalse ?
*        cmp.w   uwidth(a6,[up].w),d2 ;;
*        ble.s   findwdt ?            ;;
*        move.w  uwidth(a6,[up].w),d2 ;;
*findwdt lea     1(a0,d2.w),a0    ;;add.w   d2,a0   ;;this words lfa
*                ;move.b  (a0)+,d1 ;
*                ;cmp.b   (a0),d1 ;;
*                ;beq.s   findnfa ;;
*                ;subq.l  #1,a0   ;;
*;;findnfa
*        moveq   #63,d2
*        and.b   d0,d2    no precedence-bit
*        cmp.b   (a1),d2 ;;sub.b  (a1)+,d2 count- & smudge-bits
*        bne.s   findlink ?
*        cmp.b   #$81,d0 ;;
*        bne.s   findscan ?
*        cmp.b   #$a0,([w])
*        beq.s   findnx  ? ;;putfalse ?
*findscan
*        moveq   #127,d2
*        and.b   ([w])+,d2 ;;([w]),d2
*        move.b  (a1)+,d1
*        eor.b   d2,d1
*        beq.s   findchk         ?
*        cmp.b   #32,d1
*        bne.s   findlink        ?
*                         ;;or.b    d1,d2
*                         ;;sub.b   #'a',d2
*                         ;;cmp.b   #26,d2
*                         ;;bcc.s   findlink        ?
*findchk cmp.l   [w],a0   ;;tst.b   ([w])+
*        bhi.s   findscan ;;bpl.s   findscan        ?
*        addq.w  #4,[w]
*        sub.w   a6,[w]
*        move.w  [w],([sp])      pfa
*        move.w  d0,-([sp])      nfa-byte
*xone    move.w  #1,-([sp])      found
*        jmp     ([nx])
*findnx  clr.w   ([sp])
*        jmp     ([nx])

enclose create  {enclose} *+2
        move.w  ([sp])+,d0
        move.w  ([sp]),[w]
        add.l   a6,[w]
        move.w  [w],d1
        addq.w  #1,d1
encldskip cmp.b ([w])+,d0
        beq.s   encldskip       ?
        move.w  [w],-([sp])
        sub.w   d1,([sp])
encword move.b  ([w])+,d2
        beq.s   encend          ?
        cmp.b   d2,d0
        bne.s   encword         ?
encend  move.w  [w],-([sp])
        sub.w   d1,([sp])
        tst.b   d2
        beq.s   encnull         ?
        addq.w  #1,[w]
encnull sub.w   d1,[w]
        move.w  [w],-([sp])
        jmp     ([nx])

xuless  move.w  ([sp])+,d0
        cmp.w   ([sp])+,d0
hitrue  bhi.s   xone            ?
xzero   clr.w   -([sp])
        jmp     ([nx])

xgreat  move.l  ([sp]),d0
        swap    d0
        move.l  d0,([sp])
xless   move.w  ([sp])+,d0
        cmp.w   ([sp])+,d0
        ble.s   xzero           ?
        bra.s   xone

expect  create  {expect} Xexpect
key     create  {key} Xkey
questterminal create {?terminal} Xquestterminal
emit    create  {emit} Xemit
type    create  {type} Xtype
idtype  create  {(.id)} *+2 ; charptr count -- ptr count ) to here
        move.l  ([sp]),d1
        lea     0([bp],d1.w),a0
        swap    d1
        move.w  udp([bp],[up].l),a1
        move.w  a1,d0
        add.l   [bp],a1
        move.b  d1,(a1)+
        subq.w  #1,d1
        bmi.s   idtjtp  ?
        addq.w  #1,d0
        move.w  d0,2([sp])
        moveq   #32,d0
idtst   moveq   #127,d3
        and.b   (a0)+,d3
        cmp.w   d0,d3
        bgt.s   idtchr ?
        tas     d3
idtchr  move.b  d3,(a1)+
        dbra    d1,idtst
idtjtp  bra     Xtype   -

domove  movem.w ([sp])+,d1/a0-a1  ( from to n --- )
fillent add.l   a6,a0
        add.l   a6,a1
rollent move.w  ([w]),d2
                        ;;move.w  d2,liveinstr-Z(a6)
        lsl.b   #2,d2
livedown bpl.s  domov   ;;liveent         ?
        add.w   d1,a0
        add.w   d1,a1
        lsl.w   #2,d2
        bra.s   livedown
domov   subq.w  #1,d1
        bmi.s   mfjnx
        jmp     ([w])
;liveinstr dc.w  0       <=modified by the {move}-routines
;liveent dbra    d1,liveinstr    ?
        ;jmp     ([nx])

blanks  create {blanks} *+2
        move.w  #32,-([sp])
        bra.s   filntr
erase   create {erase}  *+2
        clr.w   -([sp])
filntr  lea     mfill(pc),[w]
xfill   movem.w ([sp])+,d0-d1/a0  ( addr n byte --- )
        bra.s   fillent
* move.b d0,(a0)+ 10C0

Xroll   move.l  [sp],a1
        move.w  ([sp])+,d1
        move.w  d1,d0
        add.w   d0,d0
        move.w  0(a1,d0.w),(a1)
        move.l  [sp],a0
        bra.s   rollent

* : fill swap >r over c! dup 1+ r> 1- cmove ;
fill    create  {fill} Xfill
mfill   move.b  d0,(a0)+        10c0
        dbra    d1,mfill
mfjnx   jmp     ([nx])

cmove   create  {cmove} domove
ucmov   move.b  (a1)+,(a0)+     10d9
        dbra    d1,ucmov
        jmp     ([nx])

lesscmove create {<cmove} domove
lcmov   move.b  -(a1),-(a0)     1121
        dbra    d1,lcmov
        jmp     ([nx])

* : roll >r r pick sp@ dup 2+ r> wmove drop ;
roll    create  {roll} Xroll
mroll   move.w  -(a1),-(a0)     3121
        dbra    d1,mroll
        jmp     ([nx])

mbymod  create  {m/mod} *+2 ;m/mod
        move.w  ([sp])+,d0 dvs
        beq.s   dropmone ?
        moveq   #0,d1      unsigned                              
        move.w  ([sp])+,d1 dvd.hi                                
        divu    d0,d1      q.hi=d1:={rmd.hi|qot.hi}=d.hi/dvs     
        move.l  d1,d2      r.hi                                  
        move.w  ([sp])+,d2 dvd.lo                                
        divu    d0,d2      d.lo  d2:={r.lo|q.lo}={r.hi|d.lo}/dvs 
        swap    d2
        move.l  d2,-([sp]) put r.lo & q.lo                       
        move.w  d1,-([sp]) q.hi
        jmp     ([nx])

umul    create  {u*} *+2
        move.w  ([sp])+,d0
        mulu    ([sp])+,d0
        bra.s   upushl

uby     create  {u/} *+2
        move.w  ([sp])+,d1
        beq.s   putmone ? ;;
        move.l  ([sp])+,d0
        divu    d1,d0
uswapsh swap    d0
upushl  move.l  d0,-([sp])
        jmp     ([nx])

dropmone addq.l #2,[sp] ;;;
putmone moveq   #-1,d0  ;;;for safety
        move.l  d0,-([sp])
        jmp     ([nx])

mmul    create  {m*} Xmmul
Xmmul   move.w  ([sp])+,d0
        muls    ([sp])+,d0
        bra.s   upushl

mby     create  {m/} *+2
        move.w  ([sp])+,d1
        beq.s   putmone ? ;;
        move.l  ([sp])+,d0
        divs    d1,d0
        bra.s   uswapsh

andx    create  {and} Xand
docand  move.w  ([w]),-([sp])
xand    move.w  ([sp])+,d0
        and.w   d0,([sp])
        jmp     ([nx])

dashtwoand dc.w docand-Z,-2
onefand dc.w    docand-Z,31

orx     create  {or} *+2
        move.w  ([sp])+,d0
        or.w    d0,([sp])
        jmp     ([nx])

xor     create  {xor} *+2
        move.w  ([sp])+,d0
        eor.w   d0,([sp])
        jmp     ([nx])

spfetch create  {sp@} Xspfetch
spstore create  {sp!} *+2
        move.w  6(a6,[up].w),[sp]
        add.l   a6,[sp]
        jmp     ([nx])

rpstore create  {rp!} Xrpstore
rpfetch create  {rp@} Xrpfetch

semis   create  {;s} *+2
        move.w  ([rp])+,[ip]
        add.l   a6,[ip]
        jmp     ([nx])

leave   create  {leave} Xleave
dleave  create  {-leave} *+2
        tst.w   ([sp])+  ;;
        beq.s   leavnx  ?;;
Xleave  move.w  ([rp]),2([rp])
leavnx  jmp     ([nx])

greatr  create  {>r} Xgreatr
greatrr create  {>rr} *+2
        move.w  ([sp]),-([sp])
xgreatr move.w  ([sp])+,-([rp])
        jmp     ([nx])

rgreat  create  {r>} *+2
        move.w  ([rp])+,-([sp])
        jmp     ([nx])

rx      create  {r} *+2
xrx     move.w  ([rp]),-([sp])
        jmp     ([nx])

ix      create  {i} Xrx
j       create  {j} *+2
        move.w  4([rp]),-([sp])
        jmp     ([nx])

nnot    create    not       *+2 ;;;additionally
        not.w     ([sp])
        jmp       ([nx])

mzeq    create  {-0=} *+2
        eor.b   #$80,([sp])
        bra.s   xzeroeq -
zeroeq  create  {0=} Xzeroeq
xeq     move.w  ([sp])+,d0
        sub.w   d0,([sp])
xzeroeq subq.w  #1,([sp])
logic   subx.w  d0,d0
        move.w  d0,([sp])
xminus  neg.w   ([sp])
        jmp     ([nx])

xabs    move.w  ([sp]),-([sp])
xplusdash tst.w ([sp])+
        blt.s   xminus          ?
        jmp     ([nx])

zeroless create {0<} *+2
        asl     ([sp])
        bra.s   logic

plus    create  {+} Xplus
docplus move.w  ([w]),-([sp])
xplus   move.w  ([sp])+,d0
        add.w   d0,([sp])
        jmp     ([nx])

dplus   create  {d+} *+2
        move.l  ([sp])+,d0
        add.l   d0,([sp])
        jmp     ([nx])

minus   create  {minus} Xminus
dminus  create  {dminus} Xdminus
xdabs   move.w  ([sp]),-([sp])
xdplusdash tst.w ([sp])+
        bge.s   nodminus        ?
xdminus neg.l   ([sp])
nodminus jmp    ([nx])

plusstore create {+!} *+2
        move.w  ([sp])+,[w]
        move.w  ([sp])+,d0
        add.w   d0,0(a6,[w].w)
        jmp     ([nx])

toggle  create  {toggle} Xtoggle
dotog   move.w  ([w]),-([sp])
xtoggle move.w  ([sp])+,d0
        move.w  ([sp])+,[w]
        eor.b   d0,0(a6,[w].w)
        jmp     ([nx])

msbtog  dc.w    dotog-Z,128
immtog  dc.w    dotog-Z,64
smutog  dc.w    dotog-Z,32

fetch   create  {@} *+2
        move.w  ([sp])+,[w]
        add.l   a6,[w]
*fetchlong
        addq.l  #1,[w]
        move.b  ([w]),-([sp])
        move.b  -([w]),-([sp])
        jmp     ([nx])

*Xfetchl
* move.l ([sp])+,[w]
* bra.s fetchlong

cfetch  create  {c@} *+2
        move.w  ([sp])+,[w]
cfetchw add.l   a6,[w]
*cfetchlong
        moveq   #0,d0
        move.b  ([w]),d0
        move.w  d0,-([sp])
        jmp     ([nx])

xcount  move.w  ([sp]),[w]
        addq.w  #1,([sp])
        bra.s   cfetchw

*Xcfetchl
* move.l ([sp])+,[w]
* bra.s cfetchlong

store   create  {!} *+2
        move.w  ([sp])+,[w]
        add.l   a6,[w]
*storelong
        move.b  ([sp])+,([w])+
        move.b  ([sp])+,([w])
        jmp     ([nx])

*Xstorel
* move.l ([sp])+,[w]
* bra.s storelong

cstore  create  {c!} *+2
        move.w  ([sp])+,[w]
        add.l   a6,[w]
*cstorelong
        move.w  ([sp])+,d0
        move.b  d0,([w])
        jmp     ([nx])

*Xcstorel
* move.l  ([sp])+,[w]
* bra.s cstorelong

twodiv  create {2/} *+2 ;signed
        asr.w   ([sp])
        jmp     ([nx])

oneplus create  {1+} Xoneplus
* : 1+ 1 + ;

onedash create  {1-} docplus
        dc.w    -1
* : 1- 1 - ;

twoplus create  {2+} *+2
        addq.w  #2,([sp])
        jmp     ([nx])
* : 2+ 2 + ;

cfa     create  {cfa} twodash+2
* : cfa 2 - ;

twodash create  {2-} *+2
        subq.w  #2,([sp])
        jmp     ([nx])
* : 2- 2 - ;

blkfetch dc.w   *+2-Z
        move.w  22(a6,[up].w),-([sp])
        jmp     ([nx])
* : blk@ blk @ ;

zeroinstore dc.w *+2-Z
        clr.w   24(a6,[up].w)
        jmp     ([nx])
* : 0in! 0 in ! ;

statefetch dc.w *+2-Z
        move.w  36(a6,[up].w),-([sp])
        jmp     ([nx])
* : state@ state @ ;

dash    create  {-} Xdash
* : - minus + ;

overdash dc.w   *+2-Z
* : over- over - ;
        move.w  2([sp]),-([sp])
xdash   move.w  ([sp])+,d0
        sub.w   d0,([sp])
        jmp     ([nx])

eq      create  {=} Xeq
* : = - 0= ;

uless   create  {u<} Xuless
less    create  {<} Xless
great   create  {>} Xgreat
* : > swap < ;

* : traverse swap begin over + 7f over c@ < until swap drop ;
traverse create {traverse} *+2
        move.w  ([sp])+,d1
travloop add.w  d1,([sp])
        move.w  ([sp]),d0
        tst.b   0(a6,d0.w)
        bpl.s   travloop        ?
        jmp     ([nx])

dashtrailing create {-trailing} *+2
        subq.w  #1,([sp])
        bmi.s   xoneplus        ?
        move.l  ([sp]),d0
        add.w   ([sp]),d0
        cmp.b   #' ',0(a6,d0.w)
        beq.s   dashtrailing+2  ?
xoneplus        
        addq.w  #1,([sp])
        jmp     ([nx])
*: -trailing dup 0 do over over + 1- c@
* bl - if leave else 1- then loop ;

basecpu dc.w    *+2-Z
        move.w  38(a6,[up].w),-([sp])
        move.w  #36,38(a6,[up].w)
        move.l  origin+8-Z(a6),-([sp])
        jmp     ([nx])

cold    create  {cold} origin

warm    create  {warm} origin+4 ;;;origin+2

sdashgreatd create {s->d} *+2
        move.w  ([sp])+,[w]
        move.l  [w],-([sp])
        jmp     ([nx])
* : s->d dup 0< minus ;

plusdash create {+-} Xplusdash
* : +- 0< if minus then ;

dplusdash create {d+-} Xdplusdash
* : d+- 0< if dminus then ;

abs     create  {abs} Xabs
* : abs dup +- ;

dabs    create  {dabs} Xdabs
* : dabs dup d+- ;

min     create  {min} *+2
        move.w  ([sp])+,d0
        cmp.w   ([sp]),d0
        bge.s   nomin           ?
mmput   move.w  d0,([sp])
nomin   jmp     ([nx])

* : min over over > if swap then drop ;
max     create  {max} *+2   ;???
        move.w  ([sp])+,d0  ;dc.w    swapx-Z
        cmp.w   ([sp]),d0   ;dc.w    min-Z
        bgt.s   mmput       ;dc.w    semis-Z
        jmp     ([nx])

*fetchl create  {@l} Xfetchl
*cfetchl create {c@l} Xcfetchl
*storel create  {!l} Xstorel
*cstorel create {c!l} Xcstorel

* qdos channels
open    create  {open} Xopen    ( name type chan --- err.l )
close   create  {close} Xclose  ( chan --- )
* keep ert at fixed(!) pan after report for "figm_scr"
report  create  {report} Xreport ( err.l chan --- )
ert     dc.w    xert-Z          ( err.l --- )
* : ert dup 0< if 1 report quit then drop drop ;

ioblo  create {io,} xioblo ( addr len --- ) ( opn-code follows like 'lit' )
temp dc.w doconstant-Z,ctemp-Z ( --- ptrstdio-1 )
* 0e constant temp

pwquoteq create {(w")} Xpwquoteq ( --- addr )
* : (w") r r @ 1+ 1 or 1+ r> + >r ;

twofetch create {2@} *+2
        move.w  ([sp])+,[w]
        add.l   a6,[w]
        move.l  ([w]),-([sp])
        jmp     ([nx])
* : 2@ dup 2+ @ swap @ ;  ( addr --- d1 )

twostore create {2!} *+2
        move.w  ([sp])+,[w]
        add.l   a6,[w]
        move.l  ([sp])+,([w])
        jmp     ([nx])
* : 2! rot over 2+ ! ! ;  ( d1 addr --- )

xpwquoteq move.w [ip],[w]
        moveq   #1,d0
        and.w   ([ip]),d0
        add.w   ([ip])+,d0
        add.w   d0,[ip]
        bra.s   pushl

* stacks *

xspfetch move.w [sp],[w]
pushl   sub.w   a6,[w]  ;;bra.s   dovariable
pushw   move.w  [w],-([sp])
        jmp     ([nx])

xrpfetch move.w [rp],[w]
        bra.s   pushl

over    create  {over} *+2
        move.w  2([sp]),-([sp])
        jmp     ([nx])

xtwoover move.l 4([sp]),-([sp])
        jmp     ([nx])

drop    create  {drop} Xdrop
xtwodrop addq.l #2,[sp]
xdrop   addq.l  #2,[sp]
        jmp     ([nx])

swapx   create  {swap} *+2
        move.l  ([sp]),d0
        swap    d0
        move.l  d0,([sp])
        jmp     ([nx])

xtwoswap move.l ([sp])+,d0
        move.l  ([sp]),d1
        move.l  d0,([sp])
        move.l  d1,-([sp])
        jmp     ([nx])

dup     create  {dup} *+2
        move.w  ([sp]),-([sp])
        jmp     ([nx])

xtwodup move.l  ([sp]),-([sp])
        jmp     ([nx])

xrot    move.l  ([sp])+,d0
        move.w  ([sp])+,[w]
        move.l  d0,-([sp])
        bra.s   pushw

rot     create  {rot} Xrot
* : rot >r swap r> swap ;

twodrop create  {2drop} Xtwodrop
* : 2drop drop drop ; ( d --- )

twodup  create  {2dup} Xtwodup
* : 2dup over over ; ( d --- d d )

twoswap create  {2swap} Xtwoswap
* : 2swap rot >r rot r> ; ( d1 d2 --- d2 d1 )

twoover create  {2over} Xtwoover
* : 2over 4 pick 4 pick ; ( d1 d2 --- d1 d2 d1 )

* : depth sp@ s0 @ swap - 2 / ;
pick    create  {pick} *+2
        move.w  ([sp]),d0
        add.w   d0,d0
        move.w  0([sp],d0.w),([sp])
        jmp     ([nx])
* : pick dup + sp@ + @ ;


dashdup create  {-dup} *+2
        move.w  ([sp]),d0
        beq.s   nodup           ?
pushd0  move.w  d0,-([sp])
nodup   jmp     ([nx])
* : -dup dup if dup then ;

depth   create  {depth} *+2
        move.w  6(a6,[up].w),d0
        add.w   a6,d0
        sub.w   [sp],d0
        asr.w   #1,d0
        bra.s   pushd0

* defining words *

colon   create  {:} docolon
        dc.w    questexec-Z
        dc.w    storecsp-Z
        dc.w    current-Z
        dc.w    fetch-Z
        dc.w    context-Z
        dc.w    store-Z
        dc.w    create-Z
        dc.w    carb-Z
        dc.w    psemicodeq-Z
previmm setnum  $140-[previmm]
do[prevlab]     
        sub.w   a6,[ip]
        move.w  [ip],-([rp])
        move.l  [w],[ip]
        jmp     ([nx])

constant create {constant} docolon
        dc.w    create-Z
        dc.w    smudge-Z
        dc.w    comma-Z
        dc.w    psemicodeq-Z
do[prevlab]     
        move.w  ([w]),-([sp])
        jmp     ([nx])

variable create {variable} docolon
        dc.w    constant-Z
        dc.w    psemicodeq-Z
do[prevlab]     
        sub.w   a6,[w]
        move.w  [w],-([sp])
        jmp     ([nx])

user    create  {user} docolon
        dc.w    constant-Z
        dc.w    psemicodeq-Z
do[prevlab]     
        move.w  ([w]),-([sp])
        add.w   [up],([sp])
        jmp     ([nx])

* compiling * +++ *

semi    create  {;} docolon
        dc.w    questcsp-Z
        dc.w    compile-Z,semis-Z ;;exit-z
        dc.w    smudge-Z
        dc.w    brac-Z
        dc.w    semis-Z
previmm setnum  $140-[previmm]

* constants & vari *

lumpplus dc.w   docplus-Z,69
dashone dc.w    doconstant-Z,-1
four    dc.w    doconstant-Z,4

mz      create  {-0} doconstant
        dc.w $8000
zero    create  {0} Xzero
one     create  {1} Xone
* 00 constant 0    01 constant 1
two     create  {2} doconstant
        dc.w    2
three   create  {3} doconstant
        dc.w    3
bl      create  {bl} doconstant
        dc.w    32
cbyl    create  {c/l} doconstant
        dc.w    64
first   create  {first} doconstant
        dc.w    0
limit   create  {limit} doconstant
        dc.w    0
bbybuf  create  {b/buf} doconstant
        dc.w    64
bbyscr  create  {b/scr} doconstant
        dc.w    16
plusorigin create {+origin} docplus
        dc.w    origin-Z

basp    create {base+} *+2
        move.l  a6,a1
        add.w   ([sp])+,a1
        move.l  a1,-([sp])
        jmp     ([nx])

us0     equ 06 user s0
ur0     equ 08 user r0
uwidth  equ 12
u0      create  {u0}  douser
        dc.w 4
s0      create  {s0}  douser
        dc.w 6
tib     create  {tib} douser
        dc.w    10
width   create  {width} douser
        dc.w    12
warning create  {warning} douser
        dc.w    14
fence   create  {fence} douser
        dc.w    16
dp      create  {dp} douser
        dc.w    18
udp     equ 18
vocdashlink create {voc-link} douser
        dc.w    20
blk     create  {blk} douser
        dc.w    22
in      create  {in} douser
        dc.w    24
out     create  {out} douser
        dc.w    26
scr     create  {scr} douser
        dc.w    28
offset  create  {offset} douser
        dc.w    30
uoffset equ 30
context create  {context} douser
        dc.w    32
ucontext equ 32
current create  {current} douser
        dc.w    34
ucurrent equ 34
state   create  {state} douser
        dc.w    36
ustate  equ 36
base    create  {base} douser
        dc.w    38
ubase   equ 38
dpl     create  {dpl} douser
        dc.w    40
fld     create  {fld} douser
        dc.w    42
csp     create  {csp} douser
        dc.w    44
rsharp  create  {r#} douser
        dc.w    46
hld     create  {hld} douser
        dc.w    48

* hi-level words *

here    create  {here} *+2
        move.w  udp(a6,[up].w),-([sp])
        jmp     ([nx])

allot   create  {allot} *+2
        move.w  ([sp])+,d0
        add.w   d0,udp(a6,[up].w)
        jmp     ([nx])

comma   create  {,} docolon
        dc.w here-Z
        dc.w two-Z
        dc.w allot-Z
        dc.w store-Z
        dc.w semis-Z

ccomma  create  {c,} docolon
        dc.w here-Z
        dc.w one-Z
        dc.w allot-Z
        dc.w cstore-Z
        dc.w semis-Z

latest  create  {latest} docolon
        dc.w    current-Z
        dc.w    fetch-Z
        dc.w    fetch-Z
        dc.w    semis-Z

lfa     create  {lfa} docplus
        dc.w    -4
* : lfa 4 - ;

fourdash equ    lfa
fiveplus dc.w   docplus-Z,5
fivedash dc.w   docplus-Z,-5

nfa     create  {nfa} docolon
        dc.w    fivedash-Z
        dc.w    dashone-Z
        dc.w    traverse-Z
        dc.w    semis-Z

pfa     create  {pfa} docolon
        dc.w    one-Z
        dc.w    traverse-Z
        dc.w    fiveplus-Z
        dc.w    semis-Z

storecsp create {!csp} docolon
        dc.w    spfetch-Z
        dc.w    csp-Z
        dc.w    store-Z
        dc.w    semis-Z

questerror create {?error} docolon
        dc.w    swapx-Z
        dc.w    zerobranch-Z
        dc.w    m1-*
        dc.w    error-Z
        dc.w    branch-Z
        dc.w    m2-*
m1      dc.w    drop-Z
m2      dc.w    semis-Z

questcomp create {?comp} docolon
        dc.w    statefetch-Z
        dc.w    zeroeq-Z
        dc.w    lit-Z,17
        dc.w    questerror-Z
        dc.w    semis-Z

questexec create {?exec} docolon
        dc.w    statefetch-Z
        dc.w    lit-Z,18
        dc.w    questerror-Z
        dc.w    semis-Z

questcsp create {?csp} docolon
        dc.w    spfetch-Z
        dc.w    csp-Z
        dc.w    fetch-Z
        dc.w    dash-Z
        dc.w    lit-Z,20
        dc.w    questerror-Z
        dc.w    semis-Z

brac    create  {`} *+2
        clr.w   ustate(a6,[up].w)
        jmp     ([nx])
previmm setnum  $140-[previmm]

carb    create  {]} *+2
        move.w  #192,ustate(a6,[up].w)
        jmp     ([nx])

compile create  {compile} docolon
        dc.w    questcomp-Z
        dc.w    rgreat-Z
        dc.w    dup-Z
        dc.w    twoplus-Z
        dc.w    greatr-Z
        dc.w    fetch-Z
        dc.w    comma-Z
        dc.w    semis-Z

smudge  create  {smudge} docolon
        dc.w    latest-Z
        dc.w    smutog-Z
        dc.w    semis-Z

decimal create  {decimal} *+2
        move.w  #10,ubase(a6,[up].w)
        jmp     ([nx])

psemicodeq create {(;code)} docolon
        dc.w    rgreat-Z
        dc.w    latest-Z
        dc.w    pfa-Z
        dc.w    cfa-Z
        dc.w    store-Z
        dc.w    semis-Z

lessbuilds create {<builds} docolon
        dc.w    zero-Z
        dc.w    constant-Z
        dc.w    semis-Z

doesgreat create {does>} docolon
        dc.w    rgreat-Z
        dc.w    latest-Z
        dc.w    pfa-Z
        dc.w    store-Z
        dc.w    psemicodeq-Z
do[prevlab]     
        sub.w   a6,[ip]
        move.w  [ip],-([rp])
        move.w  ([w])+,[ip]
        add.l   a6,[ip]
        jmp     dovariable-Z(a6)

count   create  {count} Xcount
* : count dup 1+ swap c@ ;

space   create  {space} doemit
        dc.w    32

cr      create  {cr} doemit
        dc.w    10

asciiquote dc.w doconstant-Z,0+'"'

pdotquoteq create {(.")} docolon
        dc.w    rgreat-Z
        dc.w    count-Z
        dc.w    twodup-Z
        dc.w    plus-Z
        dc.w    oneplus-Z
        dc.w    dashtwoand-Z
        dc.w    greatr-Z
        dc.w    type-Z
        dc.w    semis-Z

dotquote create {."} docolon
        dc.w    asciiquote-Z
        dc.w    statefetch-Z
        dc.w    zerobranch-Z
        dc.w    m3-*
        dc.w    compile-Z
        dc.w    pdotquoteq-Z
        dc.w    word-Z
        dc.w    here-Z
        dc.w    cfetch-Z
        dc.w    one-Z
        dc.w    orx-Z
        dc.w    oneplus-Z
        dc.w    allot-Z
        dc.w    branch-Z
        dc.w    m4-*
m3      dc.w    word-Z
        dc.w    here-Z
        dc.w    count-Z
        dc.w    type-Z
m4      dc.w    semis-Z
previmm setnum  $140-[previmm]

query   create  {query} docolon
        dc.w    tib-Z
        dc.w    fetch-Z
        dc.w    lit-Z,81
        dc.w    expect-Z
        dc.w    zeroinstore-Z
        dc.w    semis-Z

x       create  {x} docolon
        dc.w    blkfetch-Z
        dc.w    zerobranch-Z
        dc.w    m5-*
        dc.w    one-Z
        dc.w    blk-Z
        dc.w    plusstore-Z
        dc.w    zeroinstore-Z
        dc.w    blkfetch-Z
        dc.w    bbyscr-Z
        dc.w    onedash-Z
        dc.w    andx-Z
        dc.w    zeroeq-Z
        dc.w    zerobranch-Z
        dc.w    m6-*
        dc.w    questexec-Z
        dc.w    rgreat-Z
        dc.w    drop-Z
m6      dc.w    branch-Z
        dc.w    m7-*
m5      dc.w    rgreat-Z
        dc.w    drop-Z
m7      dc.w    semis-Z
previmm setnum  $140-[previmm]

hold    create  {hold} docolon
        dc.w    dashone-Z
        dc.w    hld-Z
        dc.w    plusstore-Z
        dc.w    hld-Z
        dc.w    fetch-Z
        dc.w    cstore-Z
        dc.w    semis-Z

pad     create  {pad} docolon
        dc.w    here-Z
        dc.w    lumpplus-Z
        dc.w    dashtwoand-Z
        dc.w    semis-Z

* extended to fetch from an address if blk @ = -1
word    create  {word} docolon ( { address } char --)
        dc.w    blkfetch-Z
        dc.w dashdup-Z       ;;
        dc.w    zerobranch-Z
        dc.w    m8-*
        dc.w oneplus-Z,zerobranch-Z ;;
        dc.w m8a-*   ;;
        dc.w    blkfetch-Z
        dc.w    block-Z
        dc.w    branch-Z
        dc.w    m9-*
m8a     dc.w swapx-Z,branch-Z ;;
        dc.w m9-*    ;;
m8      dc.w    tib-Z
        dc.w    fetch-Z
m9      dc.w    in-Z
        dc.w    fetch-Z
        dc.w    plus-Z
        dc.w    swapx-Z
        dc.w    enclose-Z
        dc.w    here-Z
        dc.w    asciiquote-Z
        dc.w    blanks-Z
        dc.w    in-Z
        dc.w    plusstore-Z
        dc.w    overdash-Z
        dc.w    greatrr-Z
        dc.w    here-Z
        dc.w    cstore-Z
        dc.w    plus-Z
        dc.w    here-Z
        dc.w    oneplus-Z
        dc.w    rgreat-Z
        dc.w    cmove-Z
        dc.w    semis-Z

pnumberq create {(number)} docolon
m10     dc.w    oneplus-Z
        dc.w    greatrr-Z
        dc.w    cfetch-Z
        dc.w    base-Z
        dc.w    fetch-Z
        dc.w    digit-Z
        dc.w    zerobranch-Z
        dc.w    m11-*
        dc.w    swapx-Z
        dc.w    base-Z
        dc.w    fetch-Z
        dc.w    umul-Z
        dc.w    drop-Z
        dc.w    rot-Z
        dc.w    base-Z
        dc.w    fetch-Z
        dc.w    umul-Z
        dc.w    dplus-Z
        dc.w    dpl-Z
        dc.w    fetch-Z
        dc.w    oneplus-Z
        dc.w    zerobranch-Z
        dc.w    m12-*
        dc.w    one-Z
        dc.w    dpl-Z
        dc.w    plusstore-Z
m12     dc.w    rgreat-Z
        dc.w    branch-Z
        dc.w    m10-*
m11     dc.w    rgreat-Z
        dc.w    semis-Z

number  create  {number} docolon
        dc.w    zero-Z
        dc.w    zero-Z
        dc.w    rot-Z
        dc.w    dup-Z
        dc.w    oneplus-Z
        dc.w    cfetch-Z
        dc.w    lit-Z,45
        dc.w    eq-Z
        dc.w    greatrr-Z
        dc.w    plus-Z
        dc.w    dashone-Z
m13     dc.w    dpl-Z
        dc.w    store-Z
        dc.w    pnumberq-Z
        dc.w    dup-Z
        dc.w    cfetch-Z
        dc.w    bl-Z
        dc.w    dash-Z
        dc.w    zerobranch-Z
        dc.w    m14-*
        dc.w    dup-Z
        dc.w    cfetch-Z
        dc.w    lit-Z,46
        dc.w    dash-Z
        dc.w    zero-Z
        dc.w    questerror-Z
        dc.w    zero-Z
        dc.w    branch-Z
        dc.w    m13-*
m14     dc.w    drop-Z
        dc.w    rgreat-Z
        dc.w    zerobranch-Z
        dc.w    m15-*
        dc.w    dminus-Z
m15     dc.w    semis-Z

dashfind create {-find} docolon
        dc.w bl-Z,word-Z
        dc.w here-Z,context-Z,fetch-Z
        dc.w fetch-Z,pfindq-Z
        dc.w dup-Z,zeroeq-Z,zerobranch-Z
        dc.w m16-*,drop-Z,here-Z
        dc.w latest-Z,pfindq-Z
        dc.w dup-Z,zeroeq-Z,zerobranch-Z
        dc.w m16-*,drop-Z,here-Z
        dc.w lit-Z,forthv-2-Z
        dc.w fetch-Z,pfindq-Z
m16     dc.w semis-Z

pabortq create  {(abort)} docolon
        dc.w    abort-Z
        dc.w    semis-Z

error   create  {error} docolon
        dc.w    warning-Z
        dc.w    fetch-Z
        dc.w    zeroless-Z
        dc.w    zerobranch-Z
        dc.w    m17-*
        dc.w    pabortq-Z
m17     dc.w    here-Z
        dc.w    count-Z
        dc.w    type-Z
        dc.w    pdotquoteq-Z
        dc.b    2,'?  '
        dc.w    message-Z
        dc.w    spstore-Z
        dc.w    blk-Z
        dc.w    twofetch-Z
        dc.w    quit-Z
        dc.w    semis-Z

iddot   create  {id.} docolon
        dc.w    pad-Z
        dc.w    lit-Z,32
        dc.w    lit-Z,95
        dc.w    fill-Z
        dc.w    dup-Z
        dc.w    pfa-Z
        dc.w    lfa-Z
        dc.w    overdash-Z
        dc.w    pad-Z
        dc.w    swapx-Z
        dc.w    twodup-Z
        dc.w    plus-Z
        dc.w    onedash-Z
        dc.w    greatr-Z
        dc.w    cmove-Z
        dc.w    rgreat-Z
        dc.w    msbtog-Z
        dc.w    pad-Z
        dc.w    count-Z
        dc.w    onefand-Z
        dc.w idtype-Z   ;;type-Z
        dc.w    space-Z
        dc.w    semis-Z

create  create  {create} docolon
        dc.w    dashfind-Z
        dc.w    zerobranch-Z
        dc.w    m18-*
        dc.w    drop-Z
        dc.w    nfa-Z
        dc.w    iddot-Z
        dc.w    four-Z
        dc.w    message-Z
        dc.w    space-Z
m18     dc.w    here-Z
        dc.w    dup-Z
        dc.w    count-Z
        dc.w hedchars-Z ;;clear bit 7 of chrs
        dc.w    width-Z
        dc.w    fetch-Z
        dc.w    min-Z
        dc.w    oneplus-Z
        dc.w    allot-Z
        dc.w    over-Z
        dc.w    msbtog-Z
        dc.w    here-Z
        dc.w    one-Z
        dc.w    andx-Z
        dc.w    zerobranch-Z
        dc.w    m19-*
        dc.w    one-Z
        dc.w    allot-Z
        dc.w    twodup-Z
        dc.w    here-Z
        dc.w    overdash-Z
        dc.w    lesscmove-Z
        dc.w    swapx-Z
m19     dc.w    drop-Z
        dc.w    here-Z
        dc.w    onedash-Z
        dc.w    msbtog-Z
        dc.w    latest-Z
        dc.w    comma-Z
        dc.w    current-Z
        dc.w    fetch-Z
        dc.w    store-Z
        dc.w    here-Z
        dc.w    twoplus-Z
        dc.w    comma-Z
        dc.w    smudge-Z
        dc.w    semis-Z

literal create  {literal} docolon
        dc.w    statefetch-Z
        dc.w    zerobranch-Z
        dc.w    m20-*
        dc.w    compile-Z
        dc.w    lit-Z
        dc.w    comma-Z
m20     dc.w    semis-Z
previmm setnum  $140-[previmm]

dliteral create {dliteral} docolon
        dc.w    statefetch-Z
        dc.w    zerobranch-Z
        dc.w    m21-*
        dc.w    compile-Z
        dc.w    dlit-Z
        dc.w    comma-Z
        dc.w    comma-Z
m21     dc.w    semis-Z
previmm setnum  $140-[previmm]

queststack create {?stack} docolon
        dc.w    depth-Z
        dc.w    zeroless-Z
        dc.w    one-Z
        dc.w    questerror-Z
        dc.w    spfetch-Z
        dc.w    pad-Z
        dc.w    lumpplus-Z
        dc.w    less-Z
        dc.w    lit-Z,7
        dc.w    questerror-Z
        dc.w    semis-Z

interpret create {interpret} docolon
m22     dc.w    dashfind-Z
        dc.w    zerobranch-Z
        dc.w    m23-*
        dc.w    statefetch-Z
        dc.w    less-Z
        dc.w    zerobranch-Z
        dc.w    m24-*
        dc.w    cfa-Z
        dc.w    comma-Z
        dc.w    branch-Z
        dc.w    m25-*
m24     dc.w    cfaexec-Z  ;;cfa-Z,execute-Z
m25     dc.w    queststack-Z
        dc.w    branch-Z
        dc.w    m26-*
m23     dc.w    here-Z
        dc.w    number-Z
        dc.w    dpl-Z
        dc.w    fetch-Z
        dc.w    oneplus-Z
        dc.w    zerobranch-Z
        dc.w    m27-*
        dc.w    dliteral-Z
        dc.w    branch-Z
        dc.w    m28-*
m27     dc.w    drop-Z
        dc.w    literal-Z
m28     dc.w    queststack-Z
m26     dc.w    branch-Z
        dc.w    m22-*
                ;;dc.w    semis-Z

vocabulary create {vocabulary} docolon
        dc.w    lessbuilds-Z
        dc.w    lit-Z,$81a0
        dc.w    comma-Z
        dc.w    current-Z
        dc.w    fetch-Z
        dc.w    cfa-Z
        dc.w    comma-Z
        dc.w    here-Z
        dc.w    vocdashlink-Z
        dc.w    fetch-Z
        dc.w    comma-Z
        dc.w    vocdashlink-Z
        dc.w    store-Z
        dc.w    doesgreat-Z
dovoc   dc.w    twoplus-Z
        dc.w    context-Z
        dc.w    store-Z
        dc.w    semis-Z
;;;dovoc   equ     *-8

voc setstr forth

definitions create {definitions} *+2
        move.w  ucontext(a6,[up].w),ucurrent(a6,[up].w)
        jmp     ([nx])

pquitq create {(quit)} dodefer
vpquitq dc.w quit+2-Z,quit+2-Z

loadquit dc.w docolon-Z  ;; for cmd$ with EX
        dc.w lit-Z,vpquitq-Z    ;;quit+2-Z
        dc.w defts-Z            ;;pquitq-Z,store-Z
        dc.w cr-Z,branch-Z
        dc.w m29i-*

ok      create  {ok} dodefer
        dc.w bokb+2-Z,bokb+2-Z

bokb    dc.w docolon-Z
        dc.w statefetch-Z
        dc.w zeroeq-Z
        dc.w lit-Z,60
        dc.w plus-Z
        dc.w space-Z
        dc.w emit-Z
        dc.w semis-Z

quit    create  {quit} docolon
        dc.w zero-Z
        dc.w blk-Z
        dc.w store-Z
        dc.w brac-Z
        dc.w cr-Z
m29     dc.w rpstore-Z
        dc.w ok-Z       ;;dc.w statefetch-Z
                        ;;dc.w zeroeq-Z
                        ;;dc.w lit-Z,60
                        ;;dc.w plus-Z
                        ;;dc.w space-Z
                        ;;dc.w emit-Z
        dc.w query-Z
m29i    dc.w interpret-Z
        dc.w branch-Z
        dc.w m29-*
        dc.w semis-Z

dotcpu  create  {.cpu} docolon
        dc.w    basecpu-Z
        dc.w    ddot-Z
        dc.w    base-Z
        dc.w    store-Z
        dc.w    semis-Z

abort   create  {abort} docolon
        dc.w    spstore-Z
        dc.w    drzero-Z
        dc.w    decimal-Z
        dc.w    cr-Z
        dc.w    dotcpu-Z
        dc.w    lit-Z,qfn-Z     ;;9
        dc.w    count-Z
        dc.w    type-Z
        dc.w    forth-Z
        dc.w    definitions-Z
        dc.w    pquitq-Z        ;;,perform-Z ;;quit-Z
        dc.w    semis-Z

mul     create  {*} docolon
        dc.w    umul-Z
        dc.w    drop-Z
        dc.w    semis-Z

bymod   create  {/mod} docolon
        dc.w    greatr-Z
        dc.w    sdashgreatd-Z
        dc.w    rgreat-Z
        dc.w    mby-Z
        dc.w    semis-Z

by      create  {/} docolon
        dc.w    bymod-Z
        dc.w    swapx-Z
        dc.w    drop-Z
        dc.w    semis-Z

mod     create  {mod} docolon
        dc.w    bymod-Z
        dc.w    drop-Z
        dc.w    semis-Z

mulbymod create {*/mod} docolon
        dc.w    greatr-Z
        dc.w    mmul-Z
        dc.w    rgreat-Z
        dc.w    mby-Z
        dc.w    semis-Z

mulby   create  {*/} docolon
        dc.w    mulbymod-Z
        dc.w    swapx-Z
        dc.w    drop-Z
        dc.w    semis-Z

use     create  {use} dovariable
        dc.w    0

prev    create  {prev} dovariable
        dc.w    0

plusbuf create  {+buf} docolon
        dc.w    bbybuf-Z
        dc.w    plus-Z
        dc.w    four-Z
        dc.w    plus-Z
        dc.w    dup-Z
        dc.w    limit-Z
        dc.w    eq-Z
        dc.w    zerobranch-Z
        dc.w    m30-*
        dc.w    drop-Z
        dc.w    first-Z
m30     dc.w    dup-Z
        dc.w    prev-Z
        dc.w    fetch-Z
        dc.w    dash-Z
        dc.w    semis-Z

emptydashbuffers create {empty-buffers} docolon
        dc.w    first-Z
        dc.w    limit-Z
        dc.w    overdash-Z
        dc.w    erase-Z
        dc.w    semis-Z

drzero  create  {dr0} *+2
        clr.w   uoffset(a6,[up].w)
        jmp     ([nx])

rbyw    create  {r/w} Xrbyw

buffer  create  {buffer} docolon
        dc.w    use-Z
        dc.w    fetch-Z
        dc.w    greatrr-Z
m31     dc.w    plusbuf-Z
        dc.w    zerobranch-Z
        dc.w    m31-*
        dc.w    use-Z
        dc.w    store-Z
        dc.w    rx-Z
        dc.w    fetch-Z
        dc.w    zeroless-Z
        dc.w    zerobranch-Z
        dc.w    m32-*
        dc.w    rx-Z
        dc.w    twoplus-Z
        dc.w    rx-Z
        dc.w    fetch-Z
        dc.w    lit-Z,$7fff
        dc.w    andx-Z
        dc.w    zero-Z
        dc.w    rbyw-Z
m32     dc.w    rx-Z
        dc.w    store-Z
        dc.w    rx-Z
        dc.w    prev-Z
        dc.w    store-Z
        dc.w    rgreat-Z
        dc.w    twoplus-Z
        dc.w    semis-Z

block   create  {block} docolon
        dc.w    offset-Z
        dc.w    fetch-Z
        dc.w    plus-Z
        dc.w    greatr-Z
        dc.w    prev-Z
        dc.w    fetch-Z
        dc.w    dup-Z
        dc.w    fetch-Z
        dc.w    rx-Z
        dc.w    dash-Z
        dc.w    dup-Z
        dc.w    plus-Z
        dc.w    zerobranch-Z
        dc.w    m33-*
m34     dc.w    plusbuf-Z
        dc.w    zeroeq-Z
        dc.w    zerobranch-Z
        dc.w    m35-*
        dc.w    drop-Z
        dc.w    rx-Z
        dc.w    buffer-Z
        dc.w    dup-Z
        dc.w    rx-Z
        dc.w    one-Z
        dc.w    rbyw-Z
        dc.w    twodash-Z
m35     dc.w    dup-Z
        dc.w    fetch-Z
        dc.w    rx-Z
        dc.w    dash-Z
        dc.w    dup-Z
        dc.w    plus-Z
        dc.w    zeroeq-Z
        dc.w    zerobranch-Z
        dc.w    m34-*
        dc.w    dup-Z
        dc.w    prev-Z
        dc.w    store-Z
m33     dc.w    rgreat-Z
        dc.w    drop-Z
        dc.w    twoplus-Z
        dc.w    semis-Z

plineq  create  {(line)} docolon
 dc.w    greatr-Z
 dc.w    cbyl-Z
 dc.w    bbybuf-Z
 dc.w    mulbymod-Z
 dc.w    rgreat-Z
 dc.w    bbyscr-Z
 dc.w    mul-Z
 dc.w    plus-Z
 dc.w    block-Z
 dc.w    plus-Z
 dc.w    cbyl-Z
 dc.w    semis-Z

dotline create  {.line} docolon
        dc.w    plineq-Z
        dc.w    dashtrailing-Z
        dc.w    type-Z
        dc.w    semis-Z

message create  {message} docolon
        dc.w    warning-Z
        dc.w    fetch-Z
        dc.w    zerobranch-Z
        dc.w    m36-*
        dc.w    dashdup-Z
        dc.w    zerobranch-Z
        dc.w    m37-*
        dc.w    four-Z
        dc.w    offset-Z
        dc.w    fetch-Z
        dc.w    bbyscr-Z
        dc.w    by-Z
        dc.w    dash-Z
        dc.w    dotline-Z
m37     dc.w    branch-Z
        dc.w    m38-*
m36     dc.w    pdotquoteq-Z
        dc.b    6,'msg #  '
        dc.w    dot-Z
m38     dc.w    semis-Z

load    create  {load} docolon
        dc.w    emptydashbuffers-Z ;;;ergnzt
        dc.w    blk-Z
        dc.w    twofetch-Z
        dc.w    greatr-Z
        dc.w    greatr-Z
        dc.w    zeroinstore-Z
        dc.w    bbyscr-Z
        dc.w    mul-Z
        dc.w    blk-Z
        dc.w    store-Z
        dc.w    interpret-Z
        dc.w    rgreat-Z
        dc.w    rgreat-Z
        dc.w    blk-Z
        dc.w    twostore-Z
        dc.w    semis-Z

spaces  create  {spaces} docolon
        dc.w    zero-Z
        dc.w    max-Z
        dc.w    dashdup-Z
        dc.w    zerobranch-Z
        dc.w    m43-*
        dc.w    zero-Z
        dc.w    pdoq-Z
m44     dc.w    space-Z
        dc.w    ploopq-Z
        dc.w    m44-*
m43     dc.w    semis-Z

lesssharp create {<#} docolon
        dc.w    pad-Z
        dc.w    hld-Z
        dc.w    store-Z
        dc.w    semis-Z

sharpgreat create {#>} docolon
        dc.w    twodrop-Z
        dc.w    hld-Z
        dc.w    fetch-Z
        dc.w    pad-Z
        dc.w    overdash-Z
        dc.w    semis-Z

sign    create  {sign} docolon
        dc.w    rot-Z
        dc.w    zeroless-Z
        dc.w    zerobranch-Z
        dc.w    m45-*
        dc.w    lit-Z,45
        dc.w    hold-Z
m45     dc.w    semis-Z

sharp   create  {#} docolon
        dc.w    base-Z
        dc.w    fetch-Z
        dc.w    mbymod-Z
        dc.w    rot-Z
        dc.w    lit-Z,9
        dc.w    over-Z
        dc.w    less-Z
        dc.w    zerobranch-Z
        dc.w    m46-*
        dc.w    lit-Z,7
        dc.w    plus-Z
m46     dc.w    lit-Z,48
        dc.w    plus-Z
        dc.w    hold-Z
        dc.w    semis-Z

sharps  create  {#s} docolon
m47     dc.w    sharp-Z
        dc.w    twodup-Z
        dc.w    orx-Z
        dc.w    zeroeq-Z
        dc.w    zerobranch-Z
        dc.w    m47-*
        dc.w    semis-Z

ddotr   create  {d.r} docolon
        dc.w    greatr-Z
        dc.w    swapx-Z
        dc.w    over-Z
        dc.w    dabs-Z
        dc.w    lesssharp-Z
        dc.w    sharps-Z
        dc.w    sign-Z
        dc.w    sharpgreat-Z
        dc.w    rgreat-Z
        dc.w    overdash-Z
        dc.w    spaces-Z
        dc.w    type-Z
        dc.w    semis-Z

ddot    create  {d.} docolon
        dc.w    zero-Z
        dc.w    ddotr-Z
        dc.w    space-Z
        dc.w    semis-Z

dot     create  {.} docolon
        dc.w    sdashgreatd-Z
        dc.w    ddot-Z
        dc.w    semis-Z

wquote  create  {"} docolon
 dc.w asciiquote-Z
 dc.w zero-Z
 dc.w statefetch-Z
 dc.w zerobranch-Z
 dc.w  m55-*
 dc.w compile-Z
 dc.w pwquoteq-Z
 dc.w ccomma-Z
 dc.w word-Z
 dc.w here-Z
 dc.w cfetch-Z
 dc.w oneplus-Z
 dc.w one-Z
 dc.w orx-Z
 dc.w allot-Z
 dc.w branch-Z
 dc.w  m56-*
m55
 dc.w pad-Z
 dc.w cstore-Z
 dc.w word-Z
 dc.w here-Z
 dc.w pad-Z
 dc.w oneplus-Z
 dc.w here-Z
 dc.w cfetch-Z
 dc.w oneplus-Z
 dc.w lesscmove-Z
 dc.w pad-Z
m56
 dc.w semis-Z
previmm setnum  $140-[previmm]

stdio   create  {io}    doconstant ;;extra
  dc.w cstdio-Z
savehead create {head}  doconstant ;;extra
  dc.w zsavehead-Z
task create {task} docolon end of kernal dictionary
  dc.w semis-Z

 LIST *** endit
z[prevnum] equ [previmm]
forthtop equ [forthlk]
*roottop  equ [rootlk]
voctop   equ [lastvoc]
lasttop  equ [lastlk]   ; initial {last}
;;ntop     equ prevlink]
;;top     ds.w    0
lktop    equ [prevlab]-2 ; initial {fence}
dptop    equ *-Z         ; initial {here}

dataspace equ 32764-dptop-4*chans-104-4
  data dataspace

        end

replaced by primitive:
;;mbymod  create  {m/mod} docolon
;        dc.w    greatr-Z
;        dc.w    zero-Z
;        dc.w    rx-Z
;        dc.w    uby-Z
;        dc.w    rgreat-Z
;        dc.w    swapx-Z
;        dc.w    greatr-Z
;        dc.w    uby-Z
;        dc.w    rgreat-Z
;        dc.w    semis-Z
;;erase   create  {erase} docolon
;dc.w    zero-Z
;dc.w    fill-Z
;dc.w    semis-Z
;;blanks  create  {blanks} docolon
;        dc.w    bl-Z
;        dc.w    fill-Z
;        dc.w    semis-Z
;;wmove   create  {move} domove
;uwmov   move.w  (a1)+,(a0)+     30d9
;        dbra    d1,uwmov
;        jmp     ([nx])
;*;lesswmove create {<move} domove
;lwmov   move.w  -(a1),-(a0)     3121
;        dbra    d1,lwmov
;        jmp     ([nx])
;;brac   create  {`} docolon
;dc.w    zero-Z
;dc.w    state-Z
;dc.w    store-Z
;dc.w    semis-Z
;previmm setnum  $140-[previmm]
;;carb    create  {]} docolon
;dc.w    lit-Z,192
;dc.w    state-Z
;dc.w    store-Z
;dc.w    semis-Z
;;decimal create  {decimal} docolon
;dc.w    lit-Z,10
;dc.w    base-Z
;dc.w    store-Z
;dc.w    semis-Z
;;definitions create {definitions} docolon
;dc.w    context-Z
;dc.w    fetch-Z
;dc.w    current-Z
;dc.w    store-Z
;dc.w    semis-Z
;;drzero  create  {dr0} docolon
;dc.w    zero-Z
;dc.w    offset-Z
;dc.w    store-Z
;dc.w    semis-Z
;;here    create  {here} docolon
;dc.w    dp-Z
;dc.w    fetch-Z
;dc.w    semis-Z
;;allot   create  {allot} docolon
;        dc.w    dp-Z
;        dc.w    plusstore-Z
;        dc.w    semis-Z

* additional words of the f.i.g. system to be loaded from "figm_scr":
;*;exit    create  {exit} docolon
;        dc.w    compile-Z
;        dc.w    semis-Z
;        dc.w    semis-Z
;previmm setnum  $140-[previmm]
;*;questpairs create {?pairs} docolon
;        dc.w    dash-Z
;        dc.w    lit-Z,19
;        dc.w    questerror-Z
;        dc.w    semis-Z
;*;questloading create {?loading} docolon
;        dc.w    blkfetch-Z
;        dc.w    zeroeq-Z
;        dc.w    lit-Z,22
;        dc.w    questerror-Z
;        dc.w    semis-Z
;*;hex     create  {hex} docolon
;        dc.w    lit-Z,16
;        dc.w    base-Z
;        dc.w    store-Z
;        dc.w    semis-Z
;*;semicode create {;code} docolon
;        dc.w    questcsp-Z
;        dc.w    compile-Z
;        dc.w    psemicodeq-Z
;        dc.w    brac-Z
;        dc.w    smudge-Z
;        dc.w    semis-Z
;previmm setnum  $140-[previmm]
;*;braccompilecarb create {`compile]} docolon
;        dc.w    dashfind-Z
;        dc.w    zeroeq-Z
;        dc.w    zero-Z
;        dc.w    questerror-Z
;        dc.w    drop-Z
;        dc.w    cfa-Z
;        dc.w    comma-Z
;        dc.w    semis-Z
;previmm setnum  $140-[previmm]
;*;ascii   create  {ascii} docolon
;        dc.w    bl-Z
;        dc.w    word-Z
;        dc.w    here-Z
;        dc.w    oneplus-Z
;        dc.w    cfetch-Z
;        dc.w    literal-Z
;        dc.w    semis-Z
;previmm setnum  $140-[previmm]
;*;immediate create {immediate} docolon
;        dc.w    latest-Z
;        dc.w    immtog-Z
;        dc.w    semis-Z
;*;px      create  {(} docolon
;        dc.w    lit-Z,41
;        dc.w    word-Z
;        dc.w    semis-Z
;previmm setnum  $140-[previmm]
;*;update  create  {update} docolon
;        dc.w    prev-Z
;        dc.w    fetch-Z
;        dc.w    fetch-Z
;        dc.w    lit-Z,$8000      {-32768}
;        dc.w    orx-Z
;        dc.w    prev-Z
;        dc.w    fetch-Z
;        dc.w    store-Z
;        dc.w    semis-Z
;*;drone   create  {dr1} docolon
;        dc.w    lit-Z,$4000      {16384}
;        dc.w    offset-Z
;        dc.w    store-Z
;        dc.w    semis-Z
;*;dashdashgreat create {-->} docolon
; dc.w questloading-Z
; dc.w zeroinstore-Z
; dc.w bbyscr-Z
; dc.w blkfetch-Z
; dc.w over-Z
; dc.w mod-Z
; dc.w dash-Z
; dc.w blk-Z
; dc.w plusstore-Z
; dc.w semis-Z
;previmm setnum  $140-[previmm]
;
;*;tick    create  {'} docolon
; dc.w    dashfind-Z
; dc.w    zeroeq-Z
; dc.w    zero-Z
; dc.w    questerror-Z
; dc.w    drop-Z
; dc.w    literal-Z
; dc.w    semis-Z
;previmm setnum  $140-[previmm]
;
;*;forget  create  {forget} docolon
; dc.w    tick-Z
; dc.w    nfa-Z
; dc.w    dup-Z
; dc.w    fence-Z
; dc.w    fetch-Z
; dc.w    uless-Z
; dc.w    lit-Z,21
; dc.w    questerror-Z
; dc.w    greatr-Z
; dc.w    vocdashlink-Z
; dc.w    fetch-Z
;m39 dc.w rx-Z          BEGIN
; dc.w    over-Z
; dc.w    uless-Z
; dc.w    zerobranch-Z
; dc.w m40-*            WHILE
; dc.w    forth-Z
; dc.w    definitions-Z
; dc.w    fetch-Z
; dc.w    dup-Z
; dc.w    vocdashlink-Z
; dc.w    store-Z
; dc.w    branch-Z
; dc.w    m39-*
;m40                    REPEAT
;m41 dc.w dup-Z         BEGIN
; dc.w    fourdash-Z
;m42 dc.w pfa-Z              BEGIN
; dc.w    lfa-Z
; dc.w    fetch-Z
; dc.w    dup-Z
; dc.w    rx-Z
; dc.w    uless-Z            u<
; dc.w zerobranch-Z
; dc.w m42-*                 UNTIL
; dc.w    over-Z
; dc.w    twodash-Z
; dc.w    store-Z
; dc.w    fetch-Z
; dc.w    dashdup-Z
; dc.w    zeroeq-Z      0=
; dc.w zerobranch-Z
; dc.w m41-*            UNTIL
; dc.w    rgreat-Z
; dc.w    dp-Z
; dc.w    store-Z
; dc.w    semis-Z
;
;*;back    create  {back} docolon
; dc.w    here-Z
; dc.w    dash-Z
; dc.w    comma-Z
; dc.w    semis-Z
;
;*;begin   create  {begin} docolon
; dc.w    questcomp-Z
; dc.w    here-Z
; dc.w    one-Z
; dc.w    semis-Z
;previmm setnum  $140-[previmm]
;
;*;endif   create  {endif} docolon
; dc.w    questcomp-Z
; dc.w    two-Z
; dc.w    questpairs-Z
; dc.w    here-Z
; dc.w    overdash-Z
; dc.w    swapx-Z
; dc.w    store-Z
; dc.w    semis-Z
;previmm setnum  $140-[previmm]
;
;*;then    create  {then} docolon
; dc.w    endif-Z
; dc.w    semis-Z
;previmm setnum  $140-[previmm]
;
;*;dox     create  {do} docolon
; dc.w    compile-Z
; dc.w    pdoq-Z
; dc.w    here-Z
; dc.w    three-Z
; dc.w    semis-Z
;previmm setnum  $140-[previmm]
;
;*;loop    create  {loop} docolon
; dc.w    three-Z
; dc.w    questpairs-Z
; dc.w    compile-Z
; dc.w    ploopq-Z
; dc.w    back-Z
; dc.w    semis-Z
;previmm setnum  $140-[previmm]
;
;*;plusloop create {+loop} docolon
; dc.w three-Z
; dc.w questpairs-Z
; dc.w compile-Z
; dc.w pplusloopq-Z
; dc.w back-Z
; dc.w semis-Z
;previmm setnum  $140-[previmm]
;
;*;until   create  {until} docolon
; dc.w    one-Z
; dc.w    questpairs-Z
; dc.w    compile-Z
; dc.w    zerobranch-Z
; dc.w    back-Z
; dc.w    semis-Z
;previmm setnum  $140-[previmm]
;
;*;again   create  {again} docolon
; dc.w    one-Z
; dc.w    questpairs-Z
; dc.w    compile-Z
; dc.w    branch-Z
; dc.w    back-Z
; dc.w    semis-Z
;previmm setnum  $140-[previmm]
;
;*;repeat  create  {repeat} docolon
; dc.w    twoswap-Z
; dc.w    again-Z
; dc.w    twodash-Z
; dc.w    then-Z
; dc.w    semis-Z
;previmm setnum  $140-[previmm]
;
;*;if      create  {if} docolon
; dc.w    compile-Z
; dc.w    zerobranch-Z
; dc.w    here-Z
; dc.w    zero-Z
; dc.w    comma-Z
; dc.w    two-Z
; dc.w    semis-Z
;previmm setnum  $140-[previmm]
;
;*;else    create  {else} docolon
; dc.w    two-Z
; dc.w    questpairs-Z
; dc.w    compile-Z
; dc.w    branch-Z
; dc.w    here-Z
; dc.w    zero-Z
; dc.w    comma-Z
; dc.w    swapx-Z
; dc.w    two-Z
; dc.w    then-Z
; dc.w    two-Z
; dc.w    semis-Z
;previmm setnum  $140-[previmm]
;
;*;while   create  {while} docolon
; dc.w    if-Z
; dc.w    twoplus-Z
; dc.w    semis-Z
;previmm setnum  $140-[previmm]
;*;dotr    create  {.r} docolon
; dc.w    greatr-Z
; dc.w    sdashgreatd-Z
; dc.w    rgreat-Z
; dc.w    ddotr-Z
; dc.w    semis-Z
;*;quest   create  {?} docolon
;        dc.w    fetch-Z
;        dc.w    dot-Z
;        dc.w    semis-Z
;
;*;listx   create  {list} docolon
; dc.w    decimal-Z
; dc.w    cr-Z
; dc.w    dup-Z
; dc.w    scr-Z
; dc.w    store-Z
; dc.w    pdotquoteq-Z
; dc.b    6,'scr #  '
; dc.w    dot-Z
; dc.w    lit-Z,16
; dc.w    zero-Z
; dc.w    pdoq-Z
;m48     dc.w    cr-Z
; dc.w    ix-Z
; dc.w    three-Z
; dc.w    dotr-Z
; dc.w    space-Z
; dc.w    ix-Z
; dc.w    scr-Z
; dc.w    fetch-Z
; dc.w    dotline-Z
; dc.w    ploopq-Z
; dc.w    m48-*
; dc.w    cr-Z
; dc.w    semis-Z
;
;*;index   create  {index} docolon
; dc.w    oneplus-Z
; dc.w    swapx-Z
; dc.w    pdoq-Z
;m49 dc.w    cr-Z
; dc.w    ix-Z
; dc.w    three-Z
; dc.w    dotr-Z
; dc.w    space-Z
; dc.w    zero-Z
; dc.w    ix-Z
; dc.w    dotline-Z
; dc.w    questterminal-Z
; dc.w    zerobranch-Z
; dc.w    m50-*
; dc.w    leave-Z
;m50     dc.w    ploopq-Z
; dc.w    m49-*
; dc.w    semis-Z
;
;*;triad   create  {triad} docolon
; dc.w    three-Z
; dc.w    by-Z
; dc.w    three-Z
; dc.w    mul-Z
; dc.w    three-Z
; dc.w    over-Z
; dc.w    plus-Z
; dc.w    swapx-Z
; dc.w    pdoq-Z
;m51 dc.w    cr-Z
; dc.w    ix-Z
; dc.w    listx-Z
; dc.w    ploopq-Z
; dc.w    m51-*
; dc.w    cr-Z
; dc.w    lit-Z,15
; dc.w    message-Z
; dc.w    cr-Z
; dc.w    semis-Z
;
;*;vlist   create  {vlist} docolon
; dc.w  context-Z
; dc.w  fetch-Z
; dc.w  fetch-Z
; dc.w  one-Z
;m52 dc.w zerobranch-Z
; dc.w  m53-*
; dc.w  cr-Z
; dc.w  zero-Z
; dc.w  out-Z
; dc.w  store-Z
;m53 dc.w dup-Z
; dc.w  iddot-Z
; dc.w  pfa-Z
; dc.w  lfa-Z
; dc.w  fetch-Z
; dc.w  dup-Z
; dc.w  oneplus-Z
; dc.w  cfetch-Z
; dc.w  lit-Z,160
; dc.w  eq-Z
; dc.w  questterminal-Z
; dc.w  orx-Z
; dc.w  onedash-Z
; dc.w  andx-Z
; dc.w  dashdup-Z
; dc.w  zerobranch-Z
; dc.w m54-*
; dc.w  dup-Z
; dc.w  cfetch-Z
; dc.w  onefand-Z
; dc.w  out-Z
; dc.w  fetch-Z
; dc.w  plus-Z
; dc.w  lit-Z,72
; dc.w  great-Z
; dc.w  branch-Z
; dc.w m52-*
;m54 dc.w cr-Z
; dc.w  semis-Z
;
;*;twovariable create {2variable} docolon
; dc.w    variable-Z
; dc.w    comma-Z
; dc.w    semis-Z
;
;fsheads equ 70<<3!4!1     fs.heads wait dr1
;fssave  equ 73<<3!4!1     fs.save wait dr1
;*;savequote create {save"} docolon        (save-system)
; dc.w    temp-Z
; dc.w    cfetch-Z
; dc.w    dup-Z
; dc.w    wquote-Z
; dc.w    three-Z
; dc.w    rot-Z
; dc.w    open-Z
; dc.w    ert-Z
; dc.w    lit-Z
; dc.w    savehead-Z
; dc.w    here-Z
; dc.w    over-Z
; dc.w    twoplus-Z
; dc.w    store-Z
; dc.w    zero-Z
; dc.w    ioblo-Z
; dc.w    fsheads       ;;-z
; dc.w    zero-Z
; dc.w    here-Z
; dc.w    ioblo-Z
; dc.w    fssave        ;;-z
; dc.w    close-Z
; dc.w    semis-Z
;*;chainquote create {chain"} docolon  (load from textfiles)
; dc.w    temp-Z
; dc.w    count-Z   => dr1=#15
; dc.w    swapx-Z       | #15 stdioptr #15
; dc.w    over-Z
; dc.w    wquote-Z
; dc.w    one-Z         | #15 stdio #15 name 1
; dc.w    rot-Z
; dc.w    open-Z
; dc.w    ert-Z         | #15 stdio
; dc.w    cstore-Z
; dc.w    semis-Z
;*;twoconstant create {2constant} docolon
;        dc.w    variable-Z,comma-Z
;        dc.w    psemicodeq-Z
;do[prevlab]
;        move.l  ([w]),-([sp])
;        jmp     ([nx])
